// SPDX-License-Identifier: GPL-2.0-or-later
/** @file
 * TODO: insert short description here
 *//*
 * Authors: see git history
 *
 * Copyright (C) 2018 Authors
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */
#ifndef INKSCAPE_SP_REPR_NODE_IFACE_H
#define INKSCAPE_SP_REPR_NODE_IFACE_H

#include <3rdparty/libcroco/src/cr-node-iface.h>

namespace Inkscape {
namespace XML {
extern CRNodeIface const croco_node_iface;
}
}

#endif /* !INKSCAPE_SP_REPR_NODE_IFACE_H */
