// SPDX-License-Identifier: GPL-2.0-or-later
/** @file
 * SPCSSAttr - interface for CSS Attributes
 *//*
 * Authors: see git history
 *
 * Copyright (C) 2010 Authors
 * Copyright 2005 Kees Cook <kees@outflux.net>
 * 
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */

#ifndef SEEN_INKSCAPE_XML_SP_SPCSSATTR_H
#define SEEN_INKSCAPE_XML_SP_SPCSSATTR_H

#include "xml/node.h"

class SPCSSAttr : virtual public Inkscape::XML::Node {
};

#endif
/*
  Local Variables:
  mode:c++
  c-file-style:"stroustrup"
  c-file-offsets:((innamespace . 0)(inline-open . 0)(case-label . +))
  indent-tabs-mode:nil
  fill-column:99
  End:
*/
// vim: filetype=cpp:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:fileencoding=utf-8:textwidth=99 :
