// SPDX-License-Identifier: GPL-2.0-or-later
/** @file
 * @brief Unclumping objects
 */
/* Authors:
 *   bulia byak
 *
 * Copyright (C) 2005 Authors
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */

#ifndef SEEN_DIALOGS_UNCLUMP_H
#define SEEN_DIALOGS_UNCLUMP_H

#include <vector>

class SPItem;

void unclump(std::vector<SPItem *> &items);

#endif /* !UNCLUMP_H_SEEN */

/*
  Local Variables:
  mode:c++
  c-file-style:"stroustrup"
  c-file-offsets:((innamespace . 0)(inline-open . 0)(case-label . +))
  indent-tabs-mode:nil
  fill-column:99
  End:
*/
// vim: filetype=cpp:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:fileencoding=utf-8:textwidth=99 :
