// SPDX-License-Identifier: GPL-2.0-or-later
/** @file
 * Definition of the FillOrStroke enum.
 *//*
 * Authors: see git history
 *
 * Copyright (C) 2010 Authors
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */

#ifndef SEEN_FILL_OR_STROKE_H
#define SEEN_FILL_OR_STROKE_H

/** \post STROKE == 0, FILL != 0. */
enum FillOrStroke { STROKE = 0, FILL = 1 };

#endif // !SEEN_FILL_OR_STROKE_H
