// SPDX-License-Identifier: GPL-2.0-or-later
/** @file
 * TODO: insert short description here
 *//*
 * Authors: see git history
 *
 * Copyright (C) 2014 Authors
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */
#ifndef SEEN_XML_QUOTE_H
#define SEEN_XML_QUOTE_H

#include <cstddef>

size_t xml_quoted_strlen(char const *val);
char *xml_quote_strdup(char const *src);


#endif /* !SEEN_XML_QUOTE_H */
